/*  ------------------------------------------------ */
/*  calibration   flatfield auto calib               */
/*  jfberar   berar@esrf.fr                          */
/*  Proj Res_img = Flat ( Raw_img )                  */
static char *cvs_id =" $Id: bm2cal.c,v 1.3 2004/07/29 13:54:28 berar Exp $ ";

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
/*-----------------------------------------------------------*/
#include "prg_def.h"

#include "xy_def.h"

#include "wv_def.h"
#include "gel_def.h"
#include "edf_def.h"
#include "dmi_def.h"
int cr_mode=0;

/*-----------------------------------------------------------*/
#ifndef _INCLUDE_HPUX_SOURCE
#include <time.h> 
#endif
/*-----------------------------------------------------------*/
#define TITLE "detector flatfield autocalibration"
#define RELEASE "berar@esrf.fr (d2am), "

#define FLATY 180
#define FLATX 260
#define FLATB (FLATY*750+FLATX)
#define FLATE (FLATB+5)

/* default values */
typedef struct
{   IMAGE * img;          
    float * weight;
    float scale; 
	 float added_raw_scale;           
	 int   shift_x, shift_y;
	 int   added_raw;
	 int   valid;
	 } RESULT;
typedef struct
{  IMAGE  * img;
   RESULT * result;
   float  scale;              
   float  shift_x, shift_y;
/*	int    fit_x, fit_y; */
	double sca[4];
	int    draw[4];
/*	int    dres[4]; */
	double se2, sc2, sn2;
   ZONE *excluded;
   int  n_excluded;
   int  m_excluded;
   } RAW;
IMAGE image;
/*-----------------------------------------------------------*/
/* declarations globales */
char     *prg_name;
char     *r_def = NULL;
RESULT   *current_result, *result = NULL; 
RAW      *current_raw, *raw = NULL; 
float    *flat = NULL;
int      n_result, n_raw;
T_DATA t_mode = INVALID_DATA;
float   def_scale = 1.0;

double   weight_par[5] = { 0, 0, 0, 0, 0 };
float    center_x = 0;
float    center_y = 0;
float    center_beg = 0.3;
float    center_end = 0.7;

int    set_limit = 0;
int  low_limit  = 0x80000000;
int  high_limit = 0x80000000; 
int    dformat     = 0;
int    print       = 2;
FILE *xmgr;
#if defined ( __STDC__ ) || defined ( MAC_OS )
static time_t secs_now;
#else
static long   secs_now;
#endif
char  *str_time;
/*-----------------------------------------------------------*/
FILE *inf;
/*-----------------------------------------------------------*/
void  usage( char *name);
void  parameter( FILE * f_ini);
void  put_image( char *name, char *path, IMAGE *img);
/*-----------------------------------------------------------*/
void new_result (char * buffer);
void raw_new    (char * buffer);
double res_dist (int echo);
void flat_switch (char * buffer);
void flat_new   (char * buffer);
void flat_read  (char * buffer);
void flat_cal   (char * buffer);
void flat_fit   (char * buffer);
void flat_prt   (char * buffer);
void raw_shift  (char * buffer);
void fit_shift    (char * buffer);
void center_shift    (char * buffer);
void raw_excl   (char * buffer);
void res_print  (char * buffer);
IMAGE * chk_alloc_result(RESULT * l_result);
void    build_result(int echo);
void	  add_raw(RESULT * l_result, RAW *l_raw); 
float	  raw_dist(RESULT * l_result, RAW *l_raw); 
void	  build_flat(RAW *l_raw); 
void	  norm_flat(char *name, IMAGE *img); 
char   * get_name_scale(char *n, char *b, float *s);
double   no_weight(double counts);
double   do_weight(double counts);
double  (*weight) (double counts) = no_weight;
/*-----------------------------------------------------------*/
void usage( char *name)
{
  fprintf(stderr, "usage : %s data_file\n\
   the data file has the following structure\n\
   img_files contain only one frame (dark and cosmic corrected)\n\
   see bm2cal.html for parameters\n\
   Inside the program the image are kept as integers with 32 bits\n\
   see also parameters in %s.ini file\n", prg_name, prg_name );
     exit(3);
     }
/*-----------------------------------------------------------*/
void parameter( FILE *f_ini)
{   char buffer[100], c , *d, *e;
    char *tag, *arg;
    while (!feof(f_ini))
    {   c = fgetc( f_ini );
	     ungetc(c, f_ini);
		  if (c  == '@') return;
	     fgets(buffer, sizeof(buffer)-1, f_ini);
        if (feof(f_ini)) break;
        d = buffer;
        while (*d == ' ' || *d == '\n' || *d == '\r') d++;
        if (*d=='#' || *d==0) continue;
        if ( prg_initag(d, &tag, &arg) == 0 ) continue;
        if ( xy_inipar(&image, tag, arg)) continue;
        if ( prg_iniint(tag, arg, "LOG", &log_mode)) continue;
        if ( prg_inistr(tag, arg, "DEF", &r_def)) continue;
        if ( prg_iniint(tag, arg, "PRINT", &print)) continue;
        if ( prg_iniflt(tag, arg, "SCALE", &def_scale)) continue;
        if ( prg_iniint(tag, arg, "LOW_LIMIT", &low_limit)) continue;
        if ( prg_iniint(tag, arg, "HIGH_LIMIT", &high_limit)) continue;
        if ( prg_inidbln(tag, arg, "WEIGHT", weight_par, 5)) continue;
        if ( (e=strstr(d, "T_DATA")) != 0 )
        {   e+=6;
            while (*e==' ' || *e=='=') e++;
            if (strstr(e, "T_INT32") )  t_mode = T_INT32;
  	    else if (strstr(e, "T_FLOAT") )  t_mode = T_FLOAT;
  	    else 
  	    {   prg_printf(" invalid %s mode in T_DATA, reset to INVALID_DATA\n", d);
  	        t_mode = INVALID_DATA; d = "INVALID_DATA";
  	        }
  	    prg_printf(" T_DATA= %s\n", d);
            }
       
       if (strcmp(tag, "P_EXCLUDED") == 0 )  { excl_ini( P_EXCL, arg ); continue; }
       if (strcmp(tag, "C_EXCLUDED") == 0 )  { excl_ini( C_EXCL, arg ); continue; }
       if (strcmp(tag, "R_EXCLUDED") == 0 )  { excl_ini( R_EXCL, arg ); continue; }
       if (strcmp(tag, "T_EXCLUDED") == 0 )  { excl_ini( T_EXCL, arg ); continue; }
       }
   }
/*-----------------------------------------------------------*/
int main ( int argc, char ** argv )
{   char name[200], buffer[100], *d;
    enum { USE_ARGS, ASK_NAME } mode;
    long i;

#ifdef __POWERPC__
    prg_name = "bm2cal";
    argc = 0;
#else
    prg_name = argv[0];
    argc--; argv++;
#endif    
    xy_chktype();
    r_def = ".spe";
	 if (strstr(prg_name, "spe") == prg_name) r_def = ".spe";
    if (strstr(prg_name, "edf") == prg_name) r_def = ".edf";
    if (strstr(prg_name, "gel") == prg_name) r_def = ".gel";
    memset(&image, 0, sizeof(IMAGE));
    prg_printf("## %s : %s\n## %s %s\n", prg_name, TITLE, RELEASE, __DATE__);
    mode = (argc == 0) ?  ASK_NAME : USE_ARGS;
    if ( argc == 1 ) print |= 16;
    xmgr = fopen("bm2cal.xmgr","w");
   switch (mode)
   {  case ASK_NAME :
            fprintf(stderr, "\n data file name : ");
	         buffer[0]=0;
	         fgets(buffer, sizeof(buffer)-1, stdin);
/*	         if ( redir ) puts(buffer); */
              name[0]=0;
              sscanf(buffer," %s", name);
              break;
       case USE_ARGS : strcpy(name, *argv);
             argc--; argv++;
	   }
   if ( name[0] == '-' ) usage("");
	inf = fopen(name,"r");
	if (!inf) v_error(__FILE__, "unable to open '%s' data file!", name);
	parameter(inf);
	for (i=0; i<5; i++) if (weight_par[i] != 0) weight = do_weight;
	while ( !feof(inf) )
	{  buffer[0]=0;
	   fgets(buffer, 200, inf);
      if ((d=strrchr(buffer, '\n'))) *d=0;
	   if (buffer[0]==0 || buffer[0] == '#') continue;
		if (strstr(buffer, "@COMMON") == buffer) parameter(inf);
		if (strstr(buffer, "@NEW") == buffer) new_result (buffer);
		else if (strstr(buffer, "@RAW") == buffer)   raw_new (buffer);
		else if (strstr(buffer, "@RESULT") == buffer)  res_dist (1);
		else if (strstr(buffer, "@SHIFT") == buffer) raw_shift (buffer);
		else if (strstr(buffer, "@FIT") == buffer)   fit_shift (buffer);
		else if (strstr(buffer, "@CENTER") == buffer) center_shift (buffer);
		else if (strstr(buffer, "@PRINT") == buffer) res_print(buffer);
		else if (strstr(buffer, "@FLAT") == buffer)  flat_switch(buffer);
      else if (buffer[0]=='@' && strstr(buffer, "EXCLUDED") != 0 )  raw_excl(buffer );
		else v_error(__FILE__,"unable to use statement '%s'", buffer);
		}
   PRG_EXIT
   return 0;
   }
/*-----------------------------------------------------------*/
void new_result (char * buffer)
{  char *c, *d; 
   if (result == NULL)
	{  current_result = result = (RESULT*) malloc(sizeof(RESULT));
	   n_result = 1;
		}
   else
	{  result = (RESULT*) realloc(result, (n_result+1)*sizeof(RESULT));
	   current_result = result + n_result++;
		}
	memset(current_result,0, sizeof(RESULT));
	current_result->img = (IMAGE*) prg_alloc(__FILE__, "current_result->img", sizeof(IMAGE), 1);
	current_result->img->name = get_name_scale("@NEW", buffer, &current_result->scale);
	current_result->valid = 0;
	}      
/*-----------------------------------------------------------*/
char * get_name_scale(char *name, char * buffer, float * scale)
{  char *c, *d, *e; 
	*scale = def_scale;
	c = buffer + strlen("@NEW");
	while ( *c == ' ') c++;
	e = c;
	while ( *e && *e != ' ') e++;
	*e=0;
	d = str_alloc(c);
	if (*(++e)) *scale = atof(e);
	printf ("\n %s image='%s'  scale=%.5g\n" , name, d, *scale);
	return d;
	}      
/*-----------------------------------------------------------*/
void raw_new (char * buffer)
{  char *c, *d; 
   XY_HDR *hdr;
	IMAGE  *img;
	int i, j , k;
   if (result == NULL) new_result("@NEW bm2cal ");
	if (raw == NULL)
	{  current_raw = raw = (RAW*) malloc(sizeof(RAW));
	   n_raw = 1;
		}
   else
	{  raw = (RAW*) realloc(raw, (n_raw+1)*sizeof(RAW));
	   current_raw = raw + n_raw++;
		}
	memset(current_raw,0, sizeof(RAW));
	img = current_raw->img = (IMAGE*) prg_alloc(__FILE__, "current_raw->img", sizeof(IMAGE), 1);
	memcpy(img, &image, sizeof(IMAGE));
	
	current_raw->result = current_result;
	img->name = get_name_scale("@RAW", buffer, &current_raw->scale);
	current_result->added_raw++;
	current_result->added_raw_scale += current_raw->scale;
   hdr = xy_open (img->name, prg_name, 0, r_def, 0, 1); 
   img->header = hdr;
   if (print) xy_puthdr( hdr, print & 1);
   xy_getsum(img, stdout, print, 1, 0 );
   xy_info(img, 0);
   xy_fputinfo( stdout, img);
	if (n_excluded)
	{  excl_image(img, print); 
      xy_info(img, 0);
      xy_fputinfo( stdout, img);
		}
	if (low_limit != -0x80000000) 
	for (j=0; j < img->y_size; j++)
   { for (i=0, k=j*img->x_size; i < img->x_size; i++, k++)  
	     if (img->data[k] < low_limit ) 
		  {  img->data[k] = low_limit;
		     img->frame[k] = 0;
	      }
	  }
	if (high_limit != -0x80000000) 
	for (j=0; j < img->y_size; j++)
   { for (i=0, k=j*img->x_size; i < img->x_size; i++, k++)  
	     if (img->data[k] > high_limit ) 
		  {  img->data[k] = high_limit;
		     img->frame[k] = 0;
	      }
	  }
	if ((low_limit != -0x80000000) || (high_limit != -0x80000000) )
   {  xy_info(img, 0);
      xy_fputinfo( stdout, img);
		}
   if (print & 256)
	{  printf("## map restricted to x,y<10\n");
      for (j=0; j < img->y_size && j < 10 ; j++)
      { for (i=0, k=j*img->x_size; i < img->x_size && i < 10; i++, k++)  
	       printf(" %4d %d", img->data[k], img->frame[k]);
	    printf("\n");
	  }  }
	fprintf(xmgr, "#new_raw\n@TYPE xy\n");
	for (j=0; j < img->y_size; j++)
   {    k=j*img->x_size+img->x_size/2; 
	     fprintf(xmgr, "%d %d\n", j, img->data[k]);
	  }
	fprintf(xmgr, "&\n");
	current_result->valid = 0;
	}      
/*-----------------------------------------------------------*/
double res_dist (int echo)
{  char *c; 
   XY_HDR *hdr;
	RAW * l_raw;
	IMAGE  *img;
	double se2, sc2, sn2, d;
	int i, j, k;
	se2 = sc2 = sn2 = 0;
	if (echo) printf ("\n @DIST\n");
	build_result(echo);
	for (i = 0, l_raw = raw; i < n_raw; i++, l_raw++)
	{  if (l_raw->result != current_result) continue;
	   printf(" '%s' %.1f (%.2f,%.2f) ", l_raw->img->name, l_raw->scale, l_raw->shift_x, l_raw->shift_y);
		raw_dist(current_result, l_raw);
		se2 += l_raw->se2;
		sc2 += l_raw->sc2;
		sn2 += l_raw->sn2;
	   }
	d = sqrt(se2)/sn2;
	printf(" DISTANCE  %7.5lf  %7.5lf   sums %.3lg, %.3lg, %.3lg\n", 
	   sqrt(se2/sn2), sqrt(se2/sc2), se2, sc2, sn2); 
	return d;
	}      
/*-----------------------------------------------------------*/
void  flat_switch (char * buffer)
{  char *c, *d; 
	c = buffer + strlen("@FLAT");
	while ( *c == ' ') c++;
	if ( strlen(c)==0 ) c = "NEW";
	printf ("\n @FLAT  %s\n" , c);
	if (raw==NULL) v_error(__FILE__," detector size must be known before allocating flat");
	if ( strstr(c, "NEW")) flat_new(c);
	else if ( strstr(c, "FIT")) flat_fit(c);
	else if ( strstr(c, "CALC")) flat_cal(c);
	else if ( strstr(c, "READ")) flat_read(c);
	else if ( strstr(c, "PRINT")) flat_prt(c);
	}      
/*-----------------------------------------------------------*/
void  flat_new (char * buffer)
{  int i, s; 
	if (flat) free(flat);
	flat = NULL;
	s = current_raw->img->size;
	flat = (float *) prg_alloc("xy_cal"," float flat",s*sizeof(float), 0);
	for(i=0;i<s;i++) flat[i]=1.0;
	}      
/*-----------------------------------------------------------*/
void  flat_read (char * buffer)
{  int i, size; 
   IMAGE flat_img;
	XY_HDR *hdr;
	char *c;
	if (flat) free(flat);
	flat = NULL;
	size = current_raw->img->size;
	flat = (float *) prg_alloc("xy_cal"," float flat",size*sizeof(float), 0);
	for(i=0;i<size;i++) flat[i]=1.0;
	puts (" DUMMY ROUTINE, TO BE WRITTEN");
	memcpy(&flat_img, &image, sizeof(IMAGE));
	
	c = buffer + strlen("READ");
	while ( *c == ' ') c++;
	flat_img.name = str_alloc(c);
	printf ("\n @FLAT READ image %s\n" , flat_img.name);
   hdr = xy_open (flat_img.name, prg_name, 0, r_def, 0, 1); 
   flat_img.header = hdr;
   if (print) xy_puthdr( hdr, print & 1);
   xy_getsum(&flat_img, stdout, print, 1, 0 );
	for (i=0; i < size; i++)
      flat[i]=flat_img.data[i];
	free(flat_img.data);
	norm_flat("read", &image);

	}      
/*-----------------------------------------------------------*/
void  flat_cal (char * buffer)
{  RAW * l_raw;
   int i;
	char *c = buffer + strlen("CALC");
	build_result(0);
	while ( *c==' ') c++;
   if (flat==NULL) flat_new("alloc_flat");
	if ( c == NULL)
	{  puts (" DUMMY ROUTINE, TO BE WRITTEN");
	   }
	else
	{ if ( *c == '#' ) 
	  {   i = atoi(c+1);
	      l_raw = raw +i;
	      }
	  else 
	  {  for (i = 0, l_raw = raw; i < n_raw; i++, l_raw++)
	        if (strstr(c, l_raw->img->name)==c) break;
	     if ( i >= n_raw )v_error(__FILE__," image name '%s' not found in calculating flat\n", c);
		  }
	   build_flat(l_raw); 
	   current_result->valid = 0;
		res_dist(0);
		}	  
	}      
/*-----------------------------------------------------------*/
void  flat_fit (char * buffer)
{  RAW * l_raw;
   INT32  *res_val, *data;
   INT16S *fr;
	float *res_w;
   IMAGE *img, *res_img;
	float *flat_div;
	int i, j, k, l, ri, rj;
	int cycle = 0;
	double val, w, sw, dist, old, scale;
	int    r, r0, d;
	int ox, oy, x, y, s;
	int max_cycle = 5;
	double min_dr=0.01;
	char buf[20];
	char *c = buffer + strlen("FIT");
	build_result(0);
	while (sscanf(c,"%lf%n", &val, &j) == 1)
	{    if (val > 1) max_cycle = val+0.5;
	     else if (val > 0 ) min_dr = val;
		  else fprintf(stderr, " ??? '%s'\n", c);        
	     c+=j;
	     }
   if (flat==NULL) flat_new("alloc_flat");
	s = current_raw->img->size;
	flat_div = (float *) prg_alloc("xy_cal"," float flat_div",s*sizeof(float), 0);
	while (cycle++<max_cycle)
	{  memset(flat_div, 0, s*sizeof(float));
      memset(flat, 0, s*sizeof(float));
	   res_img = current_result->img;
	   res_val = res_img->data;
	   res_w  = result->weight;
	   for (l = 0, l_raw = raw; l < n_raw; l++, l_raw++)
	   {  ox = current_result->shift_x + l_raw->shift_x;
	      oy = current_result->shift_y + l_raw->shift_y;
	      img = l_raw->img;
	      data = img->data;
	      fr   = img->frame;
	      scale = current_result->scale / l_raw->scale;
         for (j=0; j < img->y_size; j++)
         {  y = j + oy;
		      for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) if (fr[k])
		      {  x = i + ox;
			      r0 = x + y * res_img->x_size;
		         val = 0;
					sw = 0;
 			      for (d=0;d<4;d++) if ( l_raw->sca[d] > 0 )
			      {  r = r0+l_raw->draw[d];
                  if ( res_w[r] <= 0) continue;
                  val += res_val[r]/res_w[r]* l_raw->sca[d];
						sw += l_raw->sca[d];
  				      }
			      if (sw) 
					{  w = weight(val/sw/scale);
					   if (w)  
					   {  flat[k]+=w * val/sw *img->data[k]*scale;
			            flat_div[k]+=w * img->data[k]*img->data[k]*scale*scale;
						}  }
if (k>=FLATB && k<FLATE)
fprintf(stderr, "%sFLAT %d %d  %.2f=%.2g/%.2g  IMG %d %d w=%.2lf val/sw=%.2lg  raw_data=%.2lg(%d) %s\n",
  (k==FLATB)?"\n":"",i,j,  flat[k]/flat_div[k], flat[k],flat_div[k], x, y, 
  w, val/sw, img->data[k]*scale, img->data[k],
  (sw && val/sw > 5 * scale)?"added":"skipped");
			      } 
		      }
         }
	   for (j=0; j < img->y_size; j++)
      {  for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) 
		     if (flat_div[k] > 0 ) flat[k]/=flat_div[k]; else flat[k] = 1.0;
         }
     sprintf(buf," cycle %d on %d", cycle, max_cycle);
	   
		  for (i=0, k=FLATY*img->x_size; i < img->x_size; i++, k++) 
if (k>=FLATB && k<FLATE) fprintf(stderr, "%sFLAT %d %d  %.5f  %.2g\n",  (k==FLATB)?"\n":"", i, FLATY, flat[k], flat_div[k]);
	  norm_flat(buf, img);
	   
		  for (i=0, k=FLATY*img->x_size; i < img->x_size; i++, k++) 
if (k>=FLATB && k<FLATE) fprintf(stderr, "%sFLAT %d %d  %.5f\n",  (k==FLATB)?"\n":"", i, FLATY, flat[k]);

	  current_result->valid = 0;
	  old = dist;
	  dist = res_dist(0);
	  if (cycle >= 2 && (old - dist) < min_dr * dist )
	  {   printf(" Fit does not improve at %lf level, break at cycle %d on %d!", min_dr, cycle, max_cycle);
	      break;
			}
	  }
	free(flat_div);
	}      
/*-----------------------------------------------------------*/
void  flat_prt (char * buffer)
{  IMAGE *img;
   INT32 *res_val;
   int i, j, k;
	char *c = buffer + strlen("PRINT");
	while ( *c==' ') c++;
	printf ("\n @FLAT PRINT %s\n" , c);
	if (flat==NULL) v_error(__FILE__," flat must be allocated / calulated first");
	img = (IMAGE*) prg_alloc(__FILE__, "flat->img", sizeof(IMAGE), 1);
	memcpy(img, &image, sizeof(IMAGE));
	img->name = str_alloc(c);
	printf (" image %s\n" , img->name);
	img->header = (XY_HDR*) prg_alloc(__FILE__, "img->header",sizeof(XY_HDR), 1);
	img->header->x_dim = current_raw->img->x_size;
	img->header->y_dim = current_raw->img->y_size;
   img->header->t_data = T_INT32;
	img->x_omax = img->y_omax = img->x_omin = img->y_omin = 0;
	xy_setsize(img);
	img->data = (INT32*) prg_alloc(__FILE__,"img->data", sizeof(INT32)*img->size, 1);
	memset (img->data, 0, img->size*sizeof(INT32));
	img->frame = (INT16S*) prg_alloc(__FILE__,"img->frame", sizeof(INT16S)*img->size, 1);
   img->used_frames = 1;
	img->header->frames=1;
   img->header->edf_disp = img->header->edf_used=0;
   xy_edfstr( img->header, "HeaderID", "EH:000001:000000:000000");
   xy_edfstr( img->header, "ByteOrder", "LowByteFirst");
   sprintf(buffer,"%hd", img->x_size);
   xy_edfstr(img->header, "Dim_1", buffer);
   sprintf(buffer,"%hd", img->y_size);
   xy_edfstr(img->header, "Dim_2", buffer);
   xy_edfstr(img->header, "Experiment", "constructed flat field");
	res_val = img->data;
	for (j=0; j < img->y_size; j++)
     for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) img->frame[k] = 1;
	for (j=0; j < img->y_size; j++)
     for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) res_val[k] = 10000*flat[k]+0.5;
	if (print)
	{  xy_puthdr( img->header, print & 1);
      xy_info(img, 0);
      xy_fputinfo( stdout, img);
		}
	fprintf(xmgr, "#print_flat\n@TYPE xy\n");
	for (j=0; j < img->y_size; j++)
   {  k=j*img->x_size+img->x_size/2; 
	   fprintf(xmgr, "%d %.2f\n", j, flat[k]);
	   }
	fprintf(xmgr, "&\n");
   put_image(c, "",img);
   xy_freeimg(img,3);
	}      
/*-----------------------------------------------------------*/
void res_print (char * buffer)
{  char *c, *d; 
   IMAGE  *img;
	INT32  *res_val;
	float  *res_sc;
	int i, j, k;
	INT32  min_value;
	c = buffer + strlen("@PRINT");
	while ( *c == ' ') c++;
	if ( strlen(c)==0 ) c = current_result->img->name;
	printf ("\n @PRINT image %s\n" , c);
	build_result(1);
	img = current_result->img;
	res_sc  = current_result->weight;
	res_val = img->data;
	min_value = low_limit * current_result->scale * current_result->added_raw / current_result->added_raw_scale;
	if (min_value > 1 || min_value <0 ) min_value--;
	for (j=0; j < img->y_size; j++)
     for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) if (res_sc[k])
		  res_val[k] = res_val[k]/res_sc[k]+0.5; else res_val[k] = min_value;
   if (print & 256)
	{  printf("## map restricted to x,y<10\n");
      for (j=0; j < img->y_size && j < 10 ; j++)
      { for (i=0, k=j*img->x_size; i < img->x_size && i < 10; i++, k++)  
	       printf(" %4d", res_val[k]);
	    printf("\n");
	  }  }
	current_result->valid = 0;
	fprintf(xmgr, "#res_print\n@TYPE xy\n");
	for (j=0; j < img->y_size; j++)
   {  k=j*img->x_size+img->x_size/2; 
	   fprintf(xmgr, "%d %d\n", j, res_val[k]);
	   }
	fprintf(xmgr, "&\n");
	put_image(c, "",current_result->img);
   }
/*-----------------------------------------------------------*/
void raw_shift (char * buffer)
{  char *c, *d; 
   XY_HDR *hdr;
	IMAGE  *img;
	int i;
	if (raw == NULL) v_error(__FILE__, " no data associated with the shift");
	c = buffer + strlen("@SHIFT");
	sscanf(buffer,"@SHIFT %f%f", &current_raw->shift_x, &current_raw->shift_y);
	printf ("\n @SHIFT image %s  shifts %8.3f  %8.3f\n" , 
	   current_raw->img->name, current_raw->shift_x, current_raw->shift_y);
	current_result->valid = 0;
   }
/*-----------------------------------------------------------*/
void fit_shift (char * buffer)
{  char *c; 
   XY_HDR *hdr;
	IMAGE  *img;
	float d, d0;
	float beg=10.0;
	float end=0.1;
	double dx, dy;
	char buf[80];
	int i, j;
	c = buffer + strlen("@FIT");
	while (*c==' ') c++;
	if ( *c == '@' ) 
	{   i = atoi(++c);
	    current_raw = raw +i;
		 while (*c && isdigit(*c)) c++;
	    }
	  else 
	  {  for (i = 0, current_raw = raw; i < n_raw; i++, current_raw++)
	        if (strstr(c, current_raw->img->name)==c) { c += strlen(current_raw->img->name); break; }
		  }
	if ( i >= n_raw )v_error(__FILE__," image name '%s' not found!\n", c);
   while ( sscanf(c, "%f%n", &d, &j) )
	{  c += j;
	   if (d > 1) beg = d; else end = d;
		}
	current_result = current_raw->result;
	printf ("\n @FIT image #%d : %s  (delta %.1f -> %f) '%s'\n starting with shifts %8.3f  %8.3f  " , 
	   i, current_raw->img->name, beg, end, c, current_raw->shift_x, current_raw->shift_y );
	if( !current_result->valid) build_result(0);
	d0 = raw_dist(current_result, current_raw); 
	while (*c)
	{   while(*c==' ') c++;
	    dx = dy = 0;
		 switch (*c)
		 { case 'x': case 'X': dx = beg; break;
		   case 'y': case 'Y': dy = beg; break;
         case  0 : continue;
			default : printf(" unvalid case '%c' : X or Y expected\n");
			}
	    while (fabs(dx) >= end || fabs(dy) > end)
	    { current_raw->shift_x += dx;
	      current_raw->shift_y += dy;
	      current_result->valid = 0;
		   printf ( " FIT '%s' delta %6.3f %6.3f, shifts %.3f %.3f  " , 
	          c, dx, dy, current_raw->shift_x, current_raw->shift_y); 
		   build_result(0);
	      d = raw_dist(current_result, current_raw); 
		  if ( d < d0 ) { d0 = d; continue; }
		  current_raw->shift_x -= dx;
	     current_raw->shift_y -= dy;
	     dx = -dx/3;
		  dy = -dy/3;
	     }
	  c++;	
	  }	
	if ( d > d0 ) 	current_result->valid=0;	
	printf (" @FIT distance=%.3lf image '%s' shifts %8.3f  %8.3f\n" , 
	   d, current_raw->img->name, current_raw->shift_x, current_raw->shift_y);
	build_result(1);
	d = raw_dist(current_result, current_raw); 
   }
/*-----------------------------------------------------------*/
int xy_com(IMAGE *img, INT32 low, INT32 high, float *xy_res);
int xy_ring(IMAGE *img, float low, float high, float *xy_res);
void center_shift (char * buffer)
{  char *c; 
	IMAGE  *img;
	INT32 beg, end;
	float cx = 0;
	float cy = 0;
	float d;
	float xy_res[5];
	char k[10];
	sscanf(buffer,"@CENTER %s %f %f", k, &cx, &cy);
	if (strstr(k,"SET"))
	{   center_beg = cx;
	    center_end = cy;
		 printf("\n @CENTER SET shifts origine using counts between %.2f and %.2f of MAX\n",
		         center_beg, center_end);
	    beg = center_beg*current_raw->img->info.maxi;
	    end = center_end*current_raw->img->info.maxi;
		 xy_com(current_raw->img, beg, end, xy_res);
		 center_x = xy_res[0];
	    center_y = xy_res[1];
	    printf (" @CENTER image '%s' shifts origine %8.3f  %8.3f\n" , 
	       current_raw->img->name, center_x, center_y);
		 d = sqrt(xy_res[2]+xy_res[4]);
		 xy_ring(current_raw->img, 0.8 * d, 1.2 * d, xy_res);
		 center_x = xy_res[0];
	    center_y = xy_res[1];
	    printf (" @CENTER image '%s' shifts origine %8.3f  %8.3f\n" , 
	       current_raw->img->name, center_x, center_y);
		 } else
	if (strstr(k,"USE"))
	{   center_x = cx - current_raw->img->x_omin;
	    center_y = cy - current_raw->img->y_omin;
		 printf("\n @CENTER USE shifts origine at %.2f and %.2f in source size,\n\
	        at %.2f %.2f in current size\n", 
		 cx ,cy, center_x, center_y);
		 } else
	if (strstr(k,"SHIFT"))
	{   printf("\n @CENTER SHIFT (origine = %.2f, %.2f, counts %.2f %.2f)\n",
	          center_x, center_y, center_beg, center_end);
	    beg = center_beg*current_raw->img->info.maxi;
	    end = center_end*current_raw->img->info.maxi;
		 xy_com(current_raw->img, beg, end, xy_res);
		 cx = center_x - xy_res[0];
		 cy = center_y - xy_res[1];
	    printf (" @CENTER image '%s' shifts %8.3f=>%8.3f  %8.3f=>%8.3f\n" , 
	       current_raw->img->name, current_raw->shift_x, cx, current_raw->shift_y, cy);
		 d = sqrt(xy_res[2]+xy_res[4]);
		 xy_ring(current_raw->img, 0.8 * d, 1.2 * d, xy_res);
	    printf (" @CENTER image '%s' shifts %8.3f=>%8.3f  %8.3f=>%8.3f\n" , 
	       current_raw->img->name, current_raw->shift_x, cx, current_raw->shift_y, cy);
		 }
	current_result->valid=0;	
   }
/*-----------------------------------------------------------*/
int xy_com(IMAGE *img, INT32 low, INT32 high, float *xy_res)
{  long i, n, v;
   double sv, sx, sy;
   double sxx, sxy, syy;
	double ix , iy;
   INT32  *val;
   val = img->data;
    
	n = 0;
	sv = sx = sy = sxx = sxy = syy = 0;
   
   for (i=0; i<img->size; i++) 
	{  v = xy_valcor(img,i);
		if ( v >= low && v <= high )
		{   n++;
		    sv += v;
			 ix = i % img->x_size ;
			 iy = i / img->x_size ;
			 sx += ix * v;
			 sy += iy * v;
			 sxx += ix * ix * v;
			 sxy += ix * iy * v;
			 syy += iy * iy * v;
			 }
		}		
	 xy_res[0] = sx / sv;
	 xy_res[1] = sy / sv; 
	 xy_res[2] = sxx/sv - xy_res[0] * xy_res[0];
	 xy_res[3] = syy/sv - xy_res[1] * xy_res[1];
	 xy_res[4] = sxy/sv - xy_res[0] * xy_res[1];
	 printf("xy_com  n %d sv %lf   x y  %f %f , %.2f %.2f %.2f \n", 
	             n, sv, xy_res[0], xy_res[1], xy_res[2], xy_res[3], xy_res[4]);
	 return n;
    }
/*-----------------------------------------------------------*/
/* fit z=a(x-xo)^2+b(x-xo)(y-yo)+c(y-yo)^2  => xo,yo    */
int xy_ring(IMAGE *img, float low, float high, float *xy_res)
{  long i, n, v,m;
   double xo ,yo, low2, high2;
   double d, sv, sx, sy;
	int ix, ixb, ixe, iy, iyb, iye;
	INT32  *val;
   val = img->data;
    
	m = 0;
	xo = xy_res[0];
	yo = xy_res[1];
	low2 = low*low;
	high2 = high*high;
	/* somme sur anneau */
   do
	{  ixb = xo - high -1;
	   if (ixb <0 ) ixb = 0;
      ixe = xo + high +1;
	   if (ixb > img->x_size ) ixe = img->x_size;
      iyb = yo - high -1;
	   if (iyb <0 ) iyb = 0;
      iye = yo + high +1;
	   if (iye > img->y_size ) iye = img->y_size;
	   n = 0;
		sv = sx = sy = 0;
	   for (iy = iyb; iy<iye; iy++) 
	   {  i=iy*img->x_size;
	      for (ix = ixb; ix<ixe; ix++) 
	      {  d = (ix-xo)*(ix-xo)+(iy-yo)*(iy-yo);
	         if (d >= low2 && d <= high2)
	         {  v = xy_valcor(img,i+ix);
		         if ( v > 0 )
		         {  n++;
		            sv += v;
			         sx += ix * v;
			         sy += iy * v;
			   }  }  }
		   }		
	    xo = sx / sv;
	    yo = sy / sv; 
	    printf("xy_ring n %d sv %lf   x y  %lf %lf \n",  n, sv, xo, yo);
	    } while (m++ < 5);
	 xy_res[0] = xo;
	 xy_res[1] = yo; 
	 printf("xy_ring n %d sv %lf   x y  %f %f \n",  n, sv, xy_res[0], xy_res[1]);
	 /* somme symetrisee */
/*   do
	{  ixb = xo - high -1;
	   if (ixb <0 ) ixb = 0;
      ixe = xo + 1;
	   if (ixb > img->x_size ) ixe = img->x_size;
      iyb = yo - high -1;
	   if (iyb <0 ) iyb = 0;
      iye = yo + 1;
	   if (iye > img->y_size ) iye = img->y_size;
	   n = 0;
		sv = sx = sy = 0;
	   for (iy = iyb; iy<iye; iy++) 
	   {  i=iy*img->x_size;
	      for (ix = ixb; ix<ixe; ix++) 
	      {  d = (ix-xo)*(ix-xo)+(iy-yo)*(iy-yo);
	         if (d >= low2 && d <= high2)
	         {  v = xy_valcor(img,i+ix);
		         if ( v > 0 )
		         {  n++;
		            sv += v;
			         sx += ix * v;
			         sy += iy * v;
			   }  }  }
		   }		
	    xo = sx / sv;
	    yo = sy / sv; 
	    printf("xy_ring n %d sv %lf   x y  %lf %lf \n",  n, sv, xo, yo);
	    } while (m++ < 5);
	 xy_res[0] = xo;
	 xy_res[1] = yo; 
	 printf("xy_ring n %d sv %lf   x y  %f %f \n",  n, sv, xy_res[0], xy_res[1]);
*/	 
	 
	 return n;
    }

/*-----------------------------------------------------------*/
void raw_excl (char * buffer)
{  ZONE *sav_excluded = NULL;
   int  sav_n_excluded = 0;
   int  sav_m_excluded = 0;
	sav_excluded = excluded;
	sav_m_excluded = m_excluded;
	sav_n_excluded = n_excluded;
	printf ("\n @RAW_EXCL '%s' : " ,current_raw->img->name);
	switch(buffer[1]) 
	{  case 'P' : excl_ini( 0, buffer+1); break;
	   case 'C' : excl_ini( 1, buffer+1); break;
	   case 'R' : excl_ini( 2, buffer+1); break;
 	   case 'T' : excl_ini( 3, buffer+1); break;
		default : v_error(__FILE__," unknown case for raw exclusion %s", buffer);
		}
	current_raw->excluded = excluded;
	current_raw->m_excluded = m_excluded;
	current_raw->n_excluded = n_excluded;
   excl_image ( current_raw->img, 1);
   xy_info(current_raw->img, 0);
   xy_fputinfo( stdout, current_raw->img);
	
	excluded = sav_excluded;
	m_excluded = sav_m_excluded;
	n_excluded = sav_n_excluded;
   }
/*-----------------------------------------------------------*/
void put_image ( char * name, char *path, IMAGE *img)
{   char *c, *n, b[200], e[200];
    int dyn = 0;
    if (img->data == NULL) V_ERROR("no data to write");
    switch ( name[0])
    {  case 'X' : dyn=0; name++; break;
       case 'Y' : dyn=1; name++; break;
       case 'Z' : dyn=2; name++; break;
       }
    if ( *name == '=' ) name++;
    prg_outname( b, name, img->name, NULL, path);
    c = img->name;
    img->name = b;
    strcpy(e, img->name);
    n = strchr(e, '.'); 
    if (n) *n=0;
    if ( strcmp(name, b) == 0 ) strcat(b,"_x");
    if ( strstr(name, ".gel") ) 
	 {   if (dyn==0) dyn=1;
	     gel_write(img, ".gel", dyn) ;
		  }
    else if ( strstr(name, ".tif") )
	 {   if (dyn==0) dyn=2;
        gel_write(img, ".tif", dyn);
       }
    else   edf_write(img, t_mode);
    img->name = c;
    }
/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/
/* void spec_image (char *name); */
/*-----------------------------------------------------------*/
IMAGE * chk_alloc_result(RESULT * l_result)
{  char *c, *d, buffer[100]; 
   XY_HDR *hdr;
	IMAGE  *img;
	long xsize = 0;
	long ysize = 0;
	RAW   *l_raw;
	float sh_xmi = 0;
	float sh_xma = 0;
	float sh_ymi = 0;
	float sh_yma = 0;
	INT32 *val;
	int i;
	for (i = 0, l_raw = raw; i < n_raw; i++, l_raw++)
	{  if (l_raw->result != l_result) continue;
	   if (l_raw->img->x_size > xsize ) xsize = l_raw->img->x_size;
	   if (l_raw->img->y_size > ysize ) ysize = l_raw->img->y_size;
		if (l_raw->shift_x > sh_xma ) sh_xma = l_raw->shift_x;
		if (l_raw->shift_x < sh_xmi ) sh_xmi = l_raw->shift_x;
		if (l_raw->shift_y > sh_yma ) sh_yma = l_raw->shift_y;
		if (l_raw->shift_y < sh_ymi ) sh_ymi = l_raw->shift_y;
	   }
	current_result->shift_x = ( sh_xmi ) ? -sh_xmi : 0;
	current_result->shift_y = ( sh_ymi ) ? -sh_ymi : 0;
	xsize += sh_xma + current_result->shift_x; 
	ysize += sh_yma + current_result->shift_y; 
	img = current_result->img;
	if (img->data && (img->x_size != xsize || img->y_size	!= ysize )) 
	{  free(img->data);
	   img->data = NULL;
	   }
	if (img->frame && (img->x_size != xsize || img->y_size	!= ysize )) 
	{  free(img->frame);
	   img->frame = NULL;
	   }
	if (result->scale && (img->x_size != xsize || img->y_size	!= ysize )) 
	{  free(result->weight);
	   result->weight = NULL;
	   }
	if ( img->header == NULL) img->header = (XY_HDR*) prg_alloc(__FILE__, "img->header",sizeof(XY_HDR), 0);
	memset(img->header, 0, sizeof(XY_HDR));
	img->header->x_dim = xsize;
	img->header->y_dim = ysize;
   img->header->t_data = T_INT32;
	img->x_omax = img->y_omax = img->x_omin = img->y_omin = 0;
	xy_setsize(img);
	if ( img->data == NULL) 
	   img->data = (INT32*) prg_alloc(__FILE__,"img->data", sizeof(INT32)*img->size, 0);
	memset (img->data, 0, img->size*sizeof(INT32));
	if ( img->frame == NULL) 
	   img->frame = (INT16S*) prg_alloc(__FILE__,"img->frame", sizeof(INT16S)*img->size, 0);
	memset (img->frame, 0, img->size*sizeof(INT16S));
	if ( result->weight == NULL) 
	  result->weight  = (float*) prg_alloc(__FILE__,"result->weight", sizeof(float)*img->size, 0);
	memset (result->weight, 0, img->size*sizeof(float));
   img->used_frames = 0;
   img->header->edf_disp = img->header->edf_used=0;
   xy_edfstr( img->header, "HeaderID", "EH:000001:000000:000000");
   xy_edfstr( img->header, "ByteOrder", "LowByteFirst");
   sprintf(buffer,"%hd", xsize);
   xy_edfstr(img->header, "Dim_1", buffer);
   sprintf(buffer,"%hd", ysize);
   xy_edfstr(img->header, "Dim_2", buffer);
   xy_edfstr(img->header, "Experiment", "bm2cal reconstructed image");
	if (print)
	{  xy_puthdr( img->header, print & 1);
      xy_info(img, 0);
      xy_fputinfo( stdout, img);
		}
   return img;
   }
/*-----------------------------------------------------------*/
void add_raw(RESULT * l_result, RAW *l_raw)
/* add data from a raw image into the resulting image */
{  INT32  *res_val, *data;
   INT16S *fr;
	float  *res_w;
   IMAGE *img, *res_img;
	int i, j, k, ri, rj, r_max;
	double val, scale;
	double acc, dx, dy;
	int    r, r0, d;
	int ox, oy, x, y;
	res_img = l_result->img;
	res_val = res_img->data;
	res_w  = result->weight;
	r_max = res_img->size;
	scale = l_result->scale / l_raw->scale;
	ox = l_result->shift_x + l_raw->shift_x;
	oy = l_result->shift_y + l_raw->shift_y;
	acc = 0.05 / scale;
	dx =  (l_result->shift_x + l_raw->shift_x ) - ox;
	dy =  (l_result->shift_y + l_raw->shift_y ) - oy;
	if ( dx < acc ) dx = 0;
	if ( dx + acc > 1) dx = 1;
	if ( dy < acc ) dy = 0;
	if ( dy + acc > 1) dy = 1;
 	l_raw->sca[0] =  (1 - dx) * (1 - dy);
	l_raw->draw[0] = 0;
   l_raw->sca[1] =  dx * (1 - dy);
	l_raw->draw[1] = 1;
   l_raw->sca[2] =  (1 - dx) * dy;
	l_raw->draw[2] = res_img->x_size;
   l_raw->sca[3] =  dx * dy;
	l_raw->draw[3] = res_img->x_size + 1;
	l_result->img->used_frames++;
	img = l_raw->img;
	data = img->data;
	fr   = img->frame;
	for (j=0; j < img->y_size; j++)
   {  y = j + oy;
		for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) if (fr[k])
		{  x = i + ox;
			r0 = x + y * res_img->x_size;
			if (flat) val = flat[k] * data[k];  else  val = data[k];
			val = val * scale + 0.5;
		   for (d=0;d<4;d++) if ( l_raw->sca[d] > 0 )
			{  r = r0+l_raw->draw[d];
			   res_val[r] += l_raw->sca[d] * val;
			   res_w[r] += l_raw->sca[d];
			   if ( r >= r_max || r <0 ) { fprintf(stderr," ERROR add_raw r=%d, r_max=%d, i=%d, j=%d,d=%d '%s\n",
				   r, r_max,i, j, d , l_raw->img->name);
               }
/*if (j >= DBGYB && j<DBGYE && i >=DBGXB && i <DBGXE)
		printf("  x %d y %d ro %d  d=%d r=%d dres[d]=%d res_val[r]=%d res_w[r]=%f sca[d]=%.2lf \n", x, y, r0, d, r, l_raw->dres[d],res_val[r],res_w[r],l_raw->sca[d]);
*/				}
			}
      } 
   }
/*-----------------------------------------------------------*/
void build_result(int echo)
{	char *c; 
   XY_HDR *hdr;
	RAW * l_raw;
	IMAGE  *img;

	int i, j, k;
   /* check if there is something to do or if result is soon known */
   if ( !current_result->valid )
	{  img = chk_alloc_result(current_result);
	   for (i = 0, l_raw = raw; i < n_raw; i++, l_raw++)
	   {  if (l_raw->result != current_result) continue;
	      add_raw(current_result, l_raw); 
	      }
	   current_result->valid = 1;	
		}
	else img = current_result->img;
   if (echo || print) 
	{  xy_prtsize(img);
	   for (j=0; j < img->y_size; j++)
        for (i=0, k=j*img->x_size; i < img->x_size; i++, k++)  
	        img->frame[k] = current_result->weight[k] + 0.5 ;
	   xy_info(img, 0);
      xy_fputinfo( stdout, img);
      if (print & 256)
	   {  printf("## map restricted to x,y<10\n");
         for (j=0; j < img->y_size && j < 10 ; j++)
         { for (i=0, k=j*img->x_size; i < img->x_size && i < 10; i++, k++)  
	          printf(" %4d %4.1f", img->data[k], current_result->weight[k]);
	        printf("\n");
	       }
         } 
		}
   }
/*-----------------------------------------------------------*/
/* calculate the distance between the result and its raw componenent */
float raw_dist(RESULT * l_result, RAW *l_raw)
{  INT32  *res_val, *data;
   INT16S *fr;
	float *res_w;
   IMAGE *img, *res_img;
	int i, j, k, ri, rj, r_max;
	double sw, val, e, c, scale;
	int    r, r0, d;
	int ox, oy, x, y;
	int notisnan=1;
	res_img = l_result->img;
	res_val = res_img->data;
	res_w  = result->weight;
	scale = l_result->scale / l_raw->scale;
	ox = l_result->shift_x + l_raw->shift_x;
	oy = l_result->shift_y + l_raw->shift_y;
	img = l_raw->img;
	r_max = res_img->size;
	data = img->data;
	fr   = img->frame;
	l_raw->se2 =  l_raw->sc2 = l_raw->sn2 = 0;
	for (j=0; j < img->y_size; j++)
   {  y = j + oy;
		for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) if (fr[k])
		{  x = i + ox;
			r0 = x + y * res_img->x_size;
		   val = 0;
			sw = 0;
			for (d=0;d<4;d++) if ( l_raw->sca[d] > 0 )
			{  r = r0+l_raw->draw[d];
			   if ( r >= r_max || r<0) fprintf(stderr," ERROR raw_dist r=%d r_max=%d, i=%d, j=%d,d=%d '%s\n",
				             r, r_max,i, j, d , l_raw->img->name);
            if ( res_w[r] <= 0) continue;
				val += res_val[r]/res_w[r] * l_raw->sca[d];
				sw += l_raw->sca[d];
				}
			if (flat) c = flat[k]*img->data[k]; else c=img->data[k]; 
      	e =  val /sw / scale - c;
			l_raw->se2 += e * e;
			l_raw->sc2 += c * c;
			l_raw->sn2 += 1;
			}
		}
 	val = sqrt(l_raw->se2/l_raw->sn2);
	printf(" distance %7.3lf  %7.5lf\n", val, sqrt(l_raw->se2/l_raw->sc2)); 
   return val;
	}
/*-----------------------------------------------------------*/
void  build_flat (RAW * l_raw)  /* to adjust one image to result */
{  INT32  *res_val, *data;
   INT16S *fr;
	float *res_w;
   IMAGE *img, *res_img;
	int i, j, k, ri, rj;
	double val, sw;
	int    r, r0, d;
	int ox, oy, x, y;
	res_img = current_result->img;
	res_val = res_img->data;
	res_w  = result->weight;
	ox = current_result->shift_x + l_raw->shift_x;
	oy = current_result->shift_y + l_raw->shift_y;
	img = l_raw->img;
	data = img->data;
	fr   = img->frame;
	for (j=0; j < img->y_size; j++)
   {  y = j + oy;
		for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) if (fr[k])
		{  x = i + ox;
			r0 = x + y * res_img->x_size;
		   val = 0;
			sw = 0;
			for (d=0;d<4;d++) if ( l_raw->sca[d] > 0 )
			{  r = r0+l_raw->draw[d];
            if ( res_w[r] )
				{  val += res_val[r]/res_w[r];
				   sw += res_w[r];
					}
				}
			if (sw) flat[k]=val/img->data[k];
			} 
      }
	norm_flat(l_raw->img->name, img);
	}
/*-----------------------------------------------------------*/
void	norm_flat(char *name, IMAGE *img)
{	double fmean, fsig;
	float  fmi, fma;
	int i,j,k;
	fmi=fma=flat[0]; fmean = fsig =0;	
	for (j=0; j < img->y_size; j++)
   {  for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) 
		 	fmean += flat[k];
      }
	fmean /= img->size;
	for (j=0; j < img->y_size; j++)
   {  for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) 
					flat[k]/=fmean;
      }
	fmean = 0;
	for (j=0; j < img->y_size; j++)
   {  for (i=0, k=j*img->x_size; i < img->x_size; i++, k++) 
		{	if (flat[k]>fma) fma =flat[k];
			if (flat[k]<fmi) fmi =flat[k];
			fmean += flat[k];
			fsig+=flat[k]*flat[k];
			}
      }
	fmean /= img->size;
	fsig = sqrt(fsig)/img->size;
	if (name) printf ("\n flat %s : ", name);
	printf ("mean  %f  sig  %f   min %f  max %f\n", fmean, fsig, fmi, fma);
   
	}      
/*-----------------------------------------------------------*/
double no_weight(double counts)
{	 return 1.0;
    }
double do_weight(double counts)
{   double w = 0;
	 double s = 0;
    double c;
	 static int dbg = 1;
	 if (counts < weight_par[0]  || counts > weight_par[1]) return 0;
	 c = counts - weight_par[0];
	 s = weight_par[4] * c * c + weight_par[3] * c + weight_par[2];
    if (dbg) { dbg = 0;
	   printf("do_weight %lf = %lf cc + %lf c + %lf,  c = %lf -%lf < %lf\n",
		weight_par[4],weight_par[3], weight_par[2], counts, weight_par[0], weight_par[1] );
		} 
			 if ( s > 0 ) return 1 / s;
	 return 1.0;
    }
/*-----------------------------------------------------------*/

